/**
 * @file fermenters.c
 * @brief Handle fermenters status
 * @author Michiel Broek <mbroek at mbse dot eu>
 *
 * Copyright (C) 2018-2022
 *
 * This file is part of the bms (Brewery Management System)
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * bms is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with ThermFerm; see the file COPYING.  If not, write to the Free
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */


#include "bms.h"
#include "xutil.h"
#include "fermenters.h"
#include "mysql.h"
#include "mqtt.h"
#include "websocket.h"
#include "nodes.h"


sys_fermenter_list	*fermenters = NULL;

extern int		debug;
extern sys_config       Config;


void fermenter_ws_send(sys_fermenter_list *fermenter)
{
    char	*msg = NULL, buf[65];

    msg = xstrcpy((char *)"{\"device\":\"fermenters\",\"node\":\"");
    msg = xstrcat(msg, fermenter->node);
    msg = xstrcat(msg, (char *)"\",\"unit\":\"");
    msg = xstrcat(msg, fermenter->alias);
    msg = xstrcat(msg, (char *)"\",\"online\":");
    msg = xstrcat(msg, fermenter->online ? (char *)"1":(char *)"0");
    msg = xstrcat(msg, (char *)",\"mode\":\"");
    msg = xstrcat(msg, fermenter->mode);
    msg = xstrcat(msg, (char *)"\",\"beeruuid\":\"");
    msg = xstrcat(msg, fermenter->beeruuid);
    msg = xstrcat(msg, (char *)"\",\"beercode\":\"");
    msg = xstrcat(msg, fermenter->beercode);
    msg = xstrcat(msg, (char *)"\",\"beername\":\"");
    msg = xstrcat(msg, fermenter->beername);
    msg = xstrcat(msg, (char *)"\",\"yeast_lo\":");
    snprintf(buf, 64, "%.3f", fermenter->yeast_lo);
    msg = xstrcat(msg, buf);
    msg = xstrcat(msg, (char *)",\"yeast_hi\":");
    snprintf(buf, 64, "%.3f", fermenter->yeast_hi);
    msg = xstrcat(msg, buf);
    if (fermenter->air_address) {
	msg = xstrcat(msg, (char *)",\"air_state\":\"");
	msg = xstrcat(msg, fermenter->air_state);
        msg = xstrcat(msg, (char *)"\",\"air_temperature\":");
        snprintf(buf, 64, "%.3f", fermenter->air_temperature);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->beer_address) {
	msg = xstrcat(msg, (char *)",\"beer_state\":\"");
	msg = xstrcat(msg, fermenter->beer_state);
        msg = xstrcat(msg, (char *)"\",\"beer_temperature\":");
        snprintf(buf, 64, "%.3f", fermenter->beer_temperature);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->chiller_address) {
	msg = xstrcat(msg, (char *)",\"chiller_state\":\"");
	msg = xstrcat(msg, fermenter->chiller_state);
        msg = xstrcat(msg, (char *)"\",\"chiller_temperature\":");
        snprintf(buf, 64, "%.3f", fermenter->chiller_temperature);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->heater_address) {
        msg = xstrcat(msg, (char *)",\"heater_state\":");
        snprintf(buf, 64, "%d", fermenter->heater_state);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->cooler_address) {
        msg = xstrcat(msg, (char *)",\"cooler_state\":");
        snprintf(buf, 64, "%d", fermenter->cooler_state);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->fan_address) {
        msg = xstrcat(msg, (char *)",\"fan_state\":");
        snprintf(buf, 64, "%d", fermenter->fan_state);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->light_address) {
	msg = xstrcat(msg, (char *)",\"light_address\":\"");
	msg = xstrcat(msg, fermenter->light_address);
        msg = xstrcat(msg, (char *)"\",\"light_state\":");
        snprintf(buf, 64, "%d", fermenter->light_state);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->door_address) {
	msg = xstrcat(msg, (char *)",\"door_address\":\"");
        msg = xstrcat(msg, fermenter->door_address);
        msg = xstrcat(msg, (char *)"\",\"door_state\":");
        snprintf(buf, 64, "%d", fermenter->door_state);
        msg = xstrcat(msg, buf);
    }
    if (fermenter->psu_address) {
        msg = xstrcat(msg, (char *)",\"psu_address\":\"");
        msg = xstrcat(msg, fermenter->psu_address);
        msg = xstrcat(msg, (char *)"\",\"psu_state\":");
        snprintf(buf, 64, "%d", fermenter->psu_state);
        msg = xstrcat(msg, buf);
    }
    msg = xstrcat(msg, (char *)",\"setpoint_low\":");
    snprintf(buf, 64, "%.3f", fermenter->setpoint_low);
    msg = xstrcat(msg, buf);
    msg = xstrcat(msg, (char *)",\"setpoint_high\":");
    snprintf(buf, 64, "%.3f", fermenter->setpoint_high);
    msg = xstrcat(msg, buf);
    msg = xstrcat(msg, (char *)",\"alarm\":");
    snprintf(buf, 64, "%d", fermenter->alarm);
    msg = xstrcat(msg, buf);
    msg = xstrcat(msg, (char *)",\"stage\":\"");
    msg = xstrcat(msg, fermenter->stage);
    msg = xstrcat(msg, (char *)"\"");

    if (fermenter->profile_uuid) {
	msg = xstrcat(msg, (char *)",\"profile_uuid\":\"");
	msg = xstrcat(msg, fermenter->profile_uuid);
	msg = xstrcat(msg, (char *)"\",\"profile_name\":\"");
	msg = xstrcat(msg, fermenter->profile_name);
	msg = xstrcat(msg, (char *)"\",\"profile_state\":\"");
	msg = xstrcat(msg, fermenter->profile_state);
	msg = xstrcat(msg, (char *)"\",\"profile_percent\":");
	snprintf(buf, 64, "%d", fermenter->profile_percent);
	msg = xstrcat(msg, buf);
	msg = xstrcat(msg, (char *)",\"profile_inittemp_high\":");
	snprintf(buf, 64, "%.3f", fermenter->profile_inittemp_high);
	msg = xstrcat(msg, buf);
	msg = xstrcat(msg, (char *)",\"profile_inittemp_low\":");
	snprintf(buf, 64, "%.3f", fermenter->profile_inittemp_low);
	msg = xstrcat(msg, buf);
	msg = xstrcat(msg, (char *)",\"profile_steps\":");
	msg = xstrcat(msg, fermenter->profile_steps);
    }

    msg = xstrcat(msg, (char *)",\"webcam_url\":\"");
    msg = xstrcat(msg, fermenter->webcam_url);
    msg = xstrcat(msg, (char *)"\",\"webcam_light\":");
    snprintf(buf, 64, "%d", fermenter->webcam_light);
    msg = xstrcat(msg, buf);
    msg = xstrcat(msg, (char *)"}");
    ws_broadcast(msg);
    free(msg);
    msg = NULL;
}



char *fermenter_paybase(void)
{
    static char *tmp;
    char	buf[33];

    tmp = xstrcpy((char *)"{\"timestamp\":");
    snprintf(buf, 32, "%ld", time(NULL));
    tmp = xstrcat(tmp, buf);
    tmp = xstrcat(tmp, (char *)",\"metric\":");
    return tmp;
}


void fermenter_ws_receive(char *payload)
{
    struct json_object  *jobj, *pobj, *iobj, *val;
    char                *node = NULL, *alias = NULL, *beeruuid = NULL, *beercode = NULL, *beername = NULL;
    char		*mode = NULL, *stage = NULL, *profile = NULL, *profile_uuid = NULL, *profile_name = NULL, *profile_steps = NULL;
    char		*topic = NULL, *pay = NULL, buf[75], *profile_command = NULL;
    float		setpoint_low = 0, setpoint_high = 0, yeast_lo = 0, yeast_hi = 0, inittemp_lo = 0, inittemp_hi = 0;
    int			heater_state = -1, cooler_state = -1, fan_state = -1, light_state = -1, profile_fridgemode = -1;

    /*
     * Process the JSON formatted payload.
     */
    jobj = json_tokener_parse(payload);
    if (json_object_object_get_ex(jobj, "node", &val))
        node = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "unit", &val))
        alias = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "beeruuid", &val))
        beeruuid = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "beercode", &val))
        beercode = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "beername", &val))
        beername = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "mode", &val))
	mode = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "stage", &val))
	stage = xstrcpy((char *)json_object_get_string(val));
    if (json_object_object_get_ex(jobj, "setpoint_low", &val))
        setpoint_low = json_object_get_double(val);
    if (json_object_object_get_ex(jobj, "setpoint_high", &val))
        setpoint_high = json_object_get_double(val);
    if (json_object_object_get_ex(jobj, "yeast_lo", &val))
        yeast_lo = json_object_get_double(val);
    if (json_object_object_get_ex(jobj, "yeast_hi", &val))
        yeast_hi = json_object_get_double(val);
    if (json_object_object_get_ex(jobj, "heater_state", &val))
	heater_state = json_object_get_int(val);
    if (json_object_object_get_ex(jobj, "cooler_state", &val))
        cooler_state = json_object_get_int(val);
    if (json_object_object_get_ex(jobj, "fan_state", &val))
        fan_state = json_object_get_int(val);
    if (json_object_object_get_ex(jobj, "light_state", &val))
	light_state = json_object_get_int(val);
    if (json_object_object_get_ex(jobj, "profile", &pobj)) {
	profile = xstrcpy((char *)json_object_get_string(pobj));
	if (profile == NULL) { // clear profile request
	    profile = xstrcpy((char *)"null");
	}
	if (json_object_object_get_ex(pobj, "uuid", &val)) {
	    profile_uuid = xstrcpy((char *)json_object_get_string(val));
	    syslog(LOG_NOTICE, "profile uuid");
	}
	if (json_object_object_get_ex(pobj, "name", &val))
	    profile_name = xstrcpy((char *)json_object_get_string(val));
	if (json_object_object_get_ex(pobj, "inittemp", &iobj)) {

	    if (json_object_object_get_ex(iobj, "low", &val))
		inittemp_lo = json_object_get_double(val);
	    if (json_object_object_get_ex(iobj, "high", &val))
		inittemp_hi = json_object_get_double(val);
	}
	if (json_object_object_get_ex(pobj, "fridgemode", &val))
	    profile_fridgemode = json_object_get_int(val);
	if (json_object_object_get_ex(pobj, "steps", &val))
	    profile_steps = xstrcpy((char *)json_object_get_string(val));
	if (json_object_object_get_ex(pobj, "command", &val)) {
	    profile_command = xstrcpy((char *)json_object_get_string(val));
	    syslog(LOG_NOTICE, "profile command %s", profile_command);
	}
    }
    json_object_put(jobj);

    /*
     * Prepare MQTT topic
     */
    topic = xstrcpy((char *)"mbv1.0/fermenters/DCMD/");
    topic = xstrcat(topic, node);
    topic = xstrcat(topic, (char *)"/");
    topic = xstrcat(topic, alias);

    if (node && alias) {
    	if (mode) {
	    syslog(LOG_NOTICE, "Set fermenter %s/%s mode %s", node, alias, mode);
	    pay = fermenter_paybase();
	    pay = xstrcat(pay, (char *)"{\"mode\":\"");
	    pay = xstrcat(pay, mode);
	    pay = xstrcat(pay, (char *)"\"}}");
	    mqtt_publish(topic, pay);
	    free(pay);
	    pay = NULL;
    	}

    	if (stage) {
	    syslog(LOG_NOTICE, "Set fermenter %s/%s stage %s", node, alias, stage);
	    pay = fermenter_paybase();
            pay = xstrcat(pay, (char *)"{\"stage\":\"");
            pay = xstrcat(pay, stage);
            pay = xstrcat(pay, (char *)"\"}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
    	}

    	if (setpoint_low > 0 && setpoint_high > 0 && setpoint_high >= setpoint_low) {
	    syslog(LOG_NOTICE, "Set fermenter %s/%s setpoint %.1f %.1f", node, alias, setpoint_low, setpoint_high);
	    pay = fermenter_paybase();
	    pay = xstrcat(pay, (char *)"{\"setpoint\":{\"low\":");
	    snprintf(buf, 64, "%.1f", setpoint_low);
	    pay = xstrcat(pay, buf);
	    pay = xstrcat(pay, (char *)",\"high\":");
	    snprintf(buf, 64, "%.1f", setpoint_high);
	    pay = xstrcat(pay, buf);
	    pay = xstrcat(pay, (char *)"}}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
    	}

        if (heater_state >= 0) {
            syslog(LOG_NOTICE, "Set fermenter %s/%s heater %d", node, alias, heater_state);
            pay = fermenter_paybase();
            pay = xstrcat(pay, (char *)"{\"heater\":{\"state\":");
	    snprintf(buf, 64, "%d", heater_state);
	    pay = xstrcat(pay, buf);
            pay = xstrcat(pay, (char *)"}}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
        }

        if (cooler_state >= 0) {
            syslog(LOG_NOTICE, "Set fermenter %s/%s cooler %d", node, alias, cooler_state);
            pay = fermenter_paybase();
            pay = xstrcat(pay, (char *)"{\"cooler\":{\"state\":");
            snprintf(buf, 64, "%d", cooler_state);
            pay = xstrcat(pay, buf);
            pay = xstrcat(pay, (char *)"}}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
        }

        if (fan_state >= 0) {
            syslog(LOG_NOTICE, "Set fermenter %s/%s fan %d", node, alias, fan_state);
            pay = fermenter_paybase();
            pay = xstrcat(pay, (char *)"{\"fan\":{\"state\":");
            snprintf(buf, 64, "%d", fan_state);
            pay = xstrcat(pay, buf);
            pay = xstrcat(pay, (char *)"}}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
        }

	if (light_state >= 0) {
	    syslog(LOG_NOTICE, "Set fermenter %s/%s light %d", node, alias, light_state);
	    pay = fermenter_paybase();
	    pay = xstrcat(pay, (char *)"{\"light\":{\"state\":");
	    snprintf(buf, 64, "%d", light_state);
	    pay = xstrcat(pay, buf);
	    pay = xstrcat(pay, (char *)"}}}");
	    mqtt_publish(topic, pay);
	    free(pay);
	    pay = NULL;
	}

	if (beeruuid && beercode && beername && (yeast_hi > yeast_lo) && (yeast_lo > 0)) {
	    syslog(LOG_NOTICE, "Set fermenter %s/%s beer %s %s", node, alias, beercode, beername);
            pay = fermenter_paybase();
            pay = xstrcat(pay, (char *)"{\"product\":{\"uuid\":\"");
	    pay = xstrcat(pay, beeruuid);
	    pay = xstrcat(pay, (char *)"\",\"code\":\"");
	    pay = xstrcat(pay, beercode);
	    pay = xstrcat(pay, (char *)"\",\"name\":\"");
	    pay = xstrcat(pay, beername);
	    pay = xstrcat(pay, (char *)"\",\"yeast_lo\":");
	    snprintf(buf, 64, "%.1f", yeast_lo);
	    pay = xstrcat(pay, buf);
	    pay = xstrcat(pay, (char *)",\"yeast_hi\":");
            snprintf(buf, 64, "%.1f", yeast_hi);
            pay = xstrcat(pay, buf);
	    pay = xstrcat(pay, (char *)"}}}");
            mqtt_publish(topic, pay);
            free(pay);
            pay = NULL;
	}

	if (profile) {
	    syslog(LOG_NOTICE, "%s", profile);
	    if (strcmp(profile, (char *)"null") == 0) {
		syslog(LOG_NOTICE, "Set fermenter %s/%s profile null", node, alias);
            	pay = fermenter_paybase();
            	pay = xstrcat(pay, (char *)"{\"profile\":null}}");
            	mqtt_publish(topic, pay);
            	free(pay);
            	pay = NULL;
	    } else if (profile_uuid && profile_name && profile_steps) {
		syslog(LOG_NOTICE, "Set fermenter %s/%s profile %s", node, alias, profile_name);
		pay = fermenter_paybase();
                pay = xstrcat(pay, (char *)"{\"profile\":{\"uuid\":\"");
		pay = xstrcat(pay, profile_uuid);
		pay = xstrcat(pay, (char *)"\",\"name\":\"");
		pay = xstrcat(pay, profile_name);
		pay = xstrcat(pay, (char *)"\",\"inittemp\":{\"low\":");
		snprintf(buf, 64, "%.1f", inittemp_lo);
		pay = xstrcat(pay, buf);
		pay = xstrcat(pay, (char *)",\"high\":");
		snprintf(buf, 64, "%.1f", inittemp_hi);
                pay = xstrcat(pay, buf);
		pay = xstrcat(pay, (char *)"},\"fridgemode\":");
		snprintf(buf, 64, "%d", profile_fridgemode);
		pay = xstrcat(pay, buf);
		pay = xstrcat(pay, (char *)",\"steps\":");
		pay = xstrcat(pay, profile_steps);
		pay = xstrcat(pay, (char *)"}}}");
		mqtt_publish(topic, pay);
                free(pay);
                pay = NULL;
	    } else if (profile_command) {
		syslog(LOG_NOTICE, "Set fermenter %s/%s profile command %s", node, alias, profile_command);
		pay = fermenter_paybase();
                pay = xstrcat(pay, (char *)"{\"profile\":{\"command\":\"");
		pay = xstrcat(pay, profile_command);
		pay = xstrcat(pay, (char *)"\"}}}");
                mqtt_publish(topic, pay);
                free(pay);
                pay = NULL;
	    }
	}
    }

    free(topic);
    if (node)
        free(node);
    if (alias)
        free(alias);
    if (beeruuid)
        free(beeruuid);
    if (beercode)
        free(beercode);
    if (beername)
        free(beername);
    if (mode)
	free(mode);
    if (stage)
	free(stage);
    if (profile)
	free(profile);
    if (profile_uuid)
	free(profile_uuid);
    if (profile_name)
	free(profile_name);
    if (profile_steps)
	free(profile_steps);
    if (profile_command)
	free(profile_command);
}



void fermenter_set(char *edge_node, char *alias, bool birth, char *payload)
{
    struct json_object	*jobj, *val, *sensor, *temp;
    sys_fermenter_list	*fermenter, *tmpp;
    bool		new_fermenter = true;

    /*
     * Search fermenter record in the memory array and use it if found.
     */
    if (fermenters) {
	for (tmpp = fermenters; tmpp; tmpp = tmpp->next) {
	    if ((strcmp(tmpp->alias, alias) == 0) && (strcmp(tmpp->node, edge_node) == 0)) {
		new_fermenter = false;
		fermenter = tmpp;
		break;
	    }
	}
    }

    if (! birth && new_fermenter) {
	printf("ERROR got DDATA and fermenter %s/%s doesn't exist\n", edge_node, alias);
	return;
    }

    /*
     * Allocate new fermenter if not yet known.
     */
    if (new_fermenter) {
	fermenter = (sys_fermenter_list *)malloc(sizeof(sys_fermenter_list));
	memset(fermenter, 0, sizeof(sys_fermenter_list));
	fermenter->alias = xstrcpy(alias);
	fermenter->node = xstrcpy(edge_node);
	fermenter->mode = xstrcpy((char *)"OFF");
	fermenter->stage = xstrcpy((char *)"PRIMARY");
    }

    fermenter->online = true;
    if (birth) {
	syslog(LOG_NOTICE, "Online fermenter %s/%s mode %s", edge_node, alias, fermenter->mode);
    }

    /*
     * Process the JSON formatted payload. 
     * Update only the fields that are found in the payload.
     */
    jobj = json_tokener_parse(payload);

    if (json_object_object_get_ex(jobj, "uuid", &val)) {
	if (fermenter->uuid)
	    free(fermenter->uuid);
	fermenter->uuid = xstrcpy((char *)json_object_get_string(val));
    }
    if (json_object_object_get_ex(jobj, "mode", &val)) {
	if (fermenter->mode) {
	    if (strcmp(fermenter->mode, (char *)json_object_get_string(val))) {
		syslog(LOG_NOTICE, "Change mode fermenter %s/%s: %s to %s", edge_node, alias, fermenter->mode, (char *)json_object_get_string(val));
	    }
	    free(fermenter->mode);
	}
	fermenter->mode = xstrcpy((char *)json_object_get_string(val));
    }
    if (json_object_object_get_ex(jobj, "stage", &val)) {
	if (fermenter->stage) {
	    if (strcmp(fermenter->stage, (char *)json_object_get_string(val))) {
		syslog(LOG_NOTICE, "Change stage fermenter %s/%s: %s to %s", edge_node, alias, fermenter->stage, (char *)json_object_get_string(val));
	    }
	    free(fermenter->stage);
	}
	fermenter->stage = xstrcpy((char *)json_object_get_string(val));
    }
    if (json_object_object_get_ex(jobj, "alarm", &val)) {
	if (fermenter->alarm != json_object_get_int(val)) {
	    syslog(LOG_NOTICE, "Change alarm fermenter %s/%s: %d to %d", edge_node, alias, fermenter->alarm, json_object_get_int(val));
	}
	fermenter->alarm = json_object_get_int(val);
    }
    if (json_object_object_get_ex(jobj, "product", &sensor)) {
	if (json_object_object_get_ex(sensor, "code", &val)) {
	    if (fermenter->beercode)
		free(fermenter->beercode);
	    fermenter->beercode = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "name", &val)) {
	    if (fermenter->beername)
		free(fermenter->beername);
	    fermenter->beername = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "uuid", &val)) {
	    if (fermenter->beeruuid)
		free(fermenter->beeruuid);
	    fermenter->beeruuid = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "yeast_lo", &val)) {
	    fermenter->yeast_lo = json_object_get_double(val);
	}
	if (json_object_object_get_ex(sensor, "yeast_hi", &val)) {
	    fermenter->yeast_hi = json_object_get_double(val);
	}
    }
    if (json_object_object_get_ex(jobj, "air", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->air_address)
		free(fermenter->air_address);
	    fermenter->air_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    if (fermenter->air_state)
		free(fermenter->air_state);
	    fermenter->air_state= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "temperature", &val)) {
	    fermenter->air_temperature = json_object_get_double(val);
	}
    }
    if (json_object_object_get_ex(jobj, "beer", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->beer_address)
		free(fermenter->beer_address);
	    fermenter->beer_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    if (fermenter->beer_state)
		free(fermenter->beer_state);
	    fermenter->beer_state= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "temperature", &val)) {
	    fermenter->beer_temperature = json_object_get_double(val);
	}
    }
    if (json_object_object_get_ex(jobj, "chiller", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->chiller_address)
		free(fermenter->chiller_address);
	    fermenter->chiller_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    if (fermenter->chiller_state)
		free(fermenter->chiller_state);
	    fermenter->chiller_state= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "temperature", &val)) {
	    fermenter->chiller_temperature = json_object_get_double(val);
	}
    }
    if (json_object_object_get_ex(jobj, "heater", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->heater_address)
		free(fermenter->heater_address);
	    fermenter->heater_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->heater_state = json_object_get_int(val);
	}
	if (json_object_object_get_ex(sensor, "usage", &val)) {
	    fermenter->heater_usage = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "cooler", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->cooler_address)
		free(fermenter->cooler_address);
	    fermenter->cooler_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->cooler_state = json_object_get_int(val);
	}
	if (json_object_object_get_ex(sensor, "usage", &val)) {
	    fermenter->cooler_usage = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "fan", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->fan_address)
		free(fermenter->fan_address);
	    fermenter->fan_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->fan_state = json_object_get_int(val);
	}
	if (json_object_object_get_ex(sensor, "usage", &val)) {
	    fermenter->fan_usage = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "light", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->light_address)
		free(fermenter->light_address);
	    fermenter->light_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->light_state = json_object_get_int(val);
	}
	if (json_object_object_get_ex(sensor, "usage", &val)) {
	    fermenter->light_usage = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "door", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->door_address)
		free(fermenter->door_address);
	    fermenter->door_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->door_state = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "psu", &sensor)) {
	if (json_object_object_get_ex(sensor, "address", &val)) {
	    if (fermenter->psu_address)
		free(fermenter->psu_address);
	    fermenter->psu_address= xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(sensor, "state", &val)) {
	    fermenter->psu_state = json_object_get_int(val);
	}
    }
    if (json_object_object_get_ex(jobj, "setpoint", &sensor)) {
	if (json_object_object_get_ex(sensor, "low", &val)) {
	    fermenter->setpoint_low = json_object_get_double(val);
	}
	if (json_object_object_get_ex(sensor, "high", &val)) {
	    fermenter->setpoint_high = json_object_get_double(val);
	}
    }
    if (json_object_object_get_ex(jobj, "profile", &sensor)) {
	if (strcmp(json_object_to_json_string_ext(sensor, 0), "null")) {
	    if (json_object_object_get_ex(sensor, "uuid", &val)) {
		if (fermenter->profile_uuid)
		    free(fermenter->profile_uuid);
		fermenter->profile_uuid = xstrcpy((char *)json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(sensor, "name", &val)) {
		if (fermenter->profile_name)
		    free(fermenter->profile_name);
		fermenter->profile_name = xstrcpy((char *)json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(sensor, "state", &val)) {
		if (fermenter->profile_state)
		    free(fermenter->profile_state);
		fermenter->profile_state = xstrcpy((char *)json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(sensor, "percent", &val)) {
		fermenter->profile_percent = json_object_get_int(val);
	    }
	    if (json_object_object_get_ex(sensor, "inittemp", &temp)) {
		if (json_object_object_get_ex(temp, "low", &val)) {
		    fermenter->profile_inittemp_low = json_object_get_double(val);
		}
            	if (json_object_object_get_ex(temp, "high", &val)) {
                    fermenter->profile_inittemp_high = json_object_get_double(val);
		}
	    }
	    if (json_object_object_get_ex(sensor, "steps", &val)) {
		if (fermenter->profile_steps)
		    free(fermenter->profile_steps);
		fermenter->profile_steps = xstrcpy((char *)json_object_to_json_string_ext(val, 0));
	    }


	} else {
	    if (fermenter->profile_uuid)
		free(fermenter->profile_uuid);
	    if (fermenter->profile_name)
		free(fermenter->profile_name);
	    if (fermenter->profile_state)
		free(fermenter->profile_state);
	    if (fermenter->profile_steps)
		free(fermenter->profile_steps);
	    fermenter->profile_uuid = fermenter->profile_name = fermenter->profile_steps = NULL;
	    fermenter->profile_state = xstrcpy((char *)"OFF");
	    fermenter->profile_percent = 0;
	    fermenter->profile_inittemp_high = fermenter->profile_inittemp_low = 0.0;
	}
    }
    json_object_put(jobj);

    fermenter_ws_send(fermenter);
//    fermenter_dump(fermenter);

    if (new_fermenter) {
    	if (fermenters == NULL) {
	    fermenters = fermenter;
	} else {
	    for (tmpp = fermenters; tmpp; tmpp = tmpp->next) {
		if (tmpp->next == NULL) {
		    tmpp->next = fermenter;
		    break;
		}
	    }
	}
	fermenter_mysql_insert(fermenter);
    } else {
	fermenter_mysql_update(fermenter);
    }
}



/*
 * With DBIRTH all active fermenters are publishd in an array.
 * With DDATA only one fermenter is published in the payload.
 */
void fermenter_birth_data(char *topic, char *payload)
{
    char		*message_type, *edge_node, *alias;
    struct json_object  *jobj, *val, *metric, *units, *unit;
    int			arraylen;

    strtok(topic, "/"); // ignore namespace
    strtok(NULL, "/");
    message_type = strtok(NULL, "/");
    edge_node = strtok(NULL, "/\0");
    alias = strtok(NULL, "/\0");

    if ((alias == NULL) && (strcmp("DBIRTH", message_type) == 0)) {
	/*
	 * Global initial DBIRTH message with array of fermenters.
	 */
	jobj = json_tokener_parse(payload);

	if (json_object_object_get_ex(jobj, "metric", &metric)) {
	    if (json_object_object_get_ex(metric, "units", &units)) {
		arraylen = json_object_array_length(units);
		for (int i = 0; i < arraylen; i++) {
		    /*
		     * Parse the array of units
		     */
		    unit = json_object_array_get_idx(units, i);

		    if (json_object_object_get_ex(unit, "alias", &val)) {
			if (alias)
			    free(alias);
			alias = xstrcpy((char *)json_object_get_string(val));
		    	fermenter_set(edge_node, alias, true, (char *)json_object_to_json_string_ext(unit, 0));
			free(alias);
			alias = NULL;
		    }
		}
	    }
	}
	json_object_put(jobj);
	return;
    }

    if (strcmp("DBIRTH", message_type) == 0) {
	/*
	 * DBIRTH for just one fermenter.
	 */
	jobj = json_tokener_parse(payload);

	if (json_object_object_get_ex(jobj, "metric", &metric)) {
	    fermenter_set(edge_node, alias, true, (char *)json_object_to_json_string_ext(metric, 0));
	}
	json_object_put(jobj);
	return;
    }

    if (strcmp("DDATA", message_type) == 0) {
	/*
	 * DDATA update messages for each fermenter.
	 */
	jobj = json_tokener_parse(payload);

	if (json_object_object_get_ex(jobj, "metric", &metric)) {
	    fermenter_set(edge_node, alias, false, (char *)json_object_to_json_string_ext(metric, 0));
	}
	json_object_put(jobj);
	return;
    }

    /*
     * The rest are errors.
     */
    printf("ERROR fermenter_birth_data: %s %s %s\n", message_type, edge_node, alias);
}



void fermenter_log(char *topic, char *payload)
{
    char                *edge_node, *alias, *line, buf[65], *logfile, *query = malloc(512);
    struct json_object  *jobj, *val, *metric, *metric2;
    fermentation_log	*log;
    bool		trigger = false;
    struct tm		*mytime;
    time_t		timestamp;
    FILE		*fp;

    strtok(topic, "/"); // ignore namespace
    strtok(NULL, "/");	// group_id
    strtok(NULL, "/");	// message_type
    edge_node = strtok(NULL, "/\0");
    alias = strtok(NULL, "/\0");

    log = (fermentation_log *)malloc(sizeof(fermentation_log));
    memset(log, 0, sizeof(fermentation_log));

    log->fermenter_node = xstrcpy(edge_node);
    log->fermenter_alias = xstrcpy(alias);
    jobj = json_tokener_parse(payload);

    if (json_object_object_get_ex(jobj, "timestamp", &val)) {
	timestamp = json_object_get_int(val);
    	log->datetime = malloc(73);
    	mytime = localtime(&timestamp);
    	snprintf(log->datetime, 73, "%04d-%02d-%02d %02d:%02d:%02d",
	    mytime->tm_year + 1900, mytime->tm_mon + 1, mytime->tm_mday, mytime->tm_hour, mytime->tm_min, mytime->tm_sec);
	if ((mytime->tm_min % 10) == 0)
	    trigger = true;
    }

    if (json_object_object_get_ex(jobj, "metric", &metric)) {

	if (json_object_object_get_ex(metric, "product", &metric2)) {
	    if (json_object_object_get_ex(metric2, "uuid", &val)) {
		if (strcmp((char *)"(null)", json_object_get_string(val)))
		    log->product_uuid = xstrcpy((char *)json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(metric2, "code", &val)) {
		if (strcmp((char *)"(null)", json_object_get_string(val)))
		    log->product_code = xstrcpy((char *)json_object_get_string(val));
	    }
	    if (json_object_object_get_ex(metric2, "name", &val)) {
		if (strcmp((char *)"(null)", json_object_get_string(val)))
		    log->product_name = xstrcpy((char *)json_object_get_string(val));
	    }
	}
	if (json_object_object_get_ex(metric, "stage", &val)) {
	    log->stage = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(metric, "mode", &val)) {
	    log->mode = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(metric, "event", &val)) {
	    if (strcmp((char *)"(null)", json_object_get_string(val)))
	    	log->event = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(metric, "fermenter_uuid", &val)) {
	    if (strcmp((char *)"(null)", json_object_get_string(val)))
	    	log->fermenter_uuid = xstrcpy((char *)json_object_get_string(val));
	}
	if (json_object_object_get_ex(metric, "temperature", &metric2)) {
	    if (json_object_object_get_ex(metric2, "air", &val)) {
		log->temperature_air = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(metric2, "beer", &val)) {
		log->temperature_beer = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(metric2, "chiller", &val)) {
		log->temperature_chiller = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(metric2, "room", &val)) {
		log->temperature_room = json_object_get_double(val);
	    }
	}
	if (json_object_object_get_ex(metric, "setpoint", &metric2)) {
	    if (json_object_object_get_ex(metric2, "low", &val)) {
		log->setpoint_low = json_object_get_double(val);
	    }
	    if (json_object_object_get_ex(metric2, "high", &val)) {
		log->setpoint_high = json_object_get_double(val);
	    }
	}
	if (json_object_object_get_ex(metric, "heater", &metric2)) {
	    if (json_object_object_get_ex(metric2, "power", &val)) {
		log->heater_power = json_object_get_int(val);
	    }
	    if (json_object_object_get_ex(metric2, "usage", &val)) {
		log->heater_usage = json_object_get_int(val);
	    }
	}
	if (json_object_object_get_ex(metric, "cooler", &metric2)) {
	    if (json_object_object_get_ex(metric2, "power", &val)) {
		log->cooler_power = json_object_get_int(val);
	    }
	    if (json_object_object_get_ex(metric2, "usage", &val)) {
		log->cooler_usage = json_object_get_int(val);
	    }
	}
	if (json_object_object_get_ex(metric, "fan", &metric2)) {
	    if (json_object_object_get_ex(metric2, "power", &val)) {
		log->fan_power = json_object_get_int(val);
	    }
	    if (json_object_object_get_ex(metric2, "usage", &val)) {
		log->fan_usage = json_object_get_int(val);
	    }
	}
    }
    json_object_put(jobj);

    /*
     * Build MySQL log.
     * Only log every 10 minutes or if something important changed.
     */
    if (log->event && strlen(log->event))
	trigger = true;
    if (trigger) {
    	snprintf(query, 511,
		"INSERT IGNORE INTO log_fermenter SET code='%s', datetime='%s', mode='%s', stage='%s', " \
		"temp_air='%.4f', temp_beer='%.4f', temp_chiller='%.4f', temp_room='%.3f', " \
		"sp_low='%.3f', sp_high='%.3f', heater_power='%d', cooler_power='%d', " \
		"event='%s', fermenter_uuid='%s'",
		log->product_code, log->datetime, log->mode, log->stage, log->temperature_air,
		log->temperature_beer, log->temperature_chiller, log->temperature_room,
		log->setpoint_low, log->setpoint_high, log->heater_power, log->cooler_power,
		(log->event) ? log->event:"", (log->fermenter_uuid) ? log->fermenter_uuid:"");
        bms_mysql_query(query);
    }
    free(query);

    /*
     * Build csv log line. Used by the web client.
     */
    line = xstrcpy(log->datetime);
    line = xstrcat(line, (char *)",");
    line = xstrcat(line, log->mode);
    line = xstrcat(line, (char *)",");
    line = xstrcat(line, log->stage);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.3f", log->temperature_air);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.3f", log->temperature_beer);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.3f", log->temperature_chiller);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.3f", log->temperature_room);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.1f", log->setpoint_low);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%.1f", log->setpoint_high);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%d", log->heater_power);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%ld", log->heater_usage);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%d", log->cooler_power);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    snprintf(buf, 64, "%ld", log->cooler_usage);
    line = xstrcat(line, buf);
    line = xstrcat(line, (char *)",");
    if (log->event)
	line = xstrcat(line, log->event);
    line = xstrcat(line, (char *)",");
    if (log->fermenter_uuid)
	line = xstrcat(line, log->fermenter_uuid);

    /*
     * Build logfile name
     */
    logfile = xstrcpy(Config.web_root);
    logfile = xstrcat(logfile, (char *)"/log/fermentation/");
    logfile = xstrcat(logfile, log->product_code);
    logfile = xstrcat(logfile, (char *)" ");
    logfile = xstrcat(logfile, log->product_name);
    logfile = xstrcat(logfile, (char *)".log");

    if (debug)
	fprintf(stdout, "%s %s\n", logfile, line);

    fp = fopen(logfile, "a");
    if (fp) {
	fprintf(fp, "%s\n", line);
	fclose(fp);
    } else {
	syslog(LOG_NOTICE, "cannot append to `%s'", logfile);
    }

    free(logfile);
    logfile = NULL;
    free(line);
    line = NULL;

    if (log->datetime)
    	free(log->datetime);
    if (log->product_uuid )
	free(log->product_uuid );
    if (log->product_code )
	free(log->product_code );
    if (log->product_name )
	free(log->product_name );
    if (log->stage)
	free(log->stage);
    if (log->mode)
	free(log->mode);
    if (log->event)
	free(log->event);
    if (log->fermenter_uuid)
	free(log->fermenter_uuid);
    if (log->fermenter_node)
	free(log->fermenter_node);
    if (log->fermenter_alias)
	free(log->fermenter_alias);
    free(log);
}



void fermenter_dump(sys_fermenter_list *fermenter)
{
    if (debug) {
    	printf("uuid      %s\n", fermenter->uuid);
    	printf("alias     %s\n", fermenter->alias);
    	printf("node      %s\n", fermenter->node);
	printf("online    %s\n", fermenter->online ? "yes":"no");
	printf("product   %s / %s\n", fermenter->beercode, fermenter->beername);
	if (fermenter->air_address) 
	    printf("Air       %-36s %10s %8.3f\n", fermenter->air_address, fermenter->air_state, fermenter->air_temperature);
	if (fermenter->beer_address) 
	    printf("Beer      %-36s %10s %8.3f\n", fermenter->beer_address, fermenter->beer_state, fermenter->beer_temperature);
	if (fermenter->chiller_address) 
	    printf("Chiller   %-36s %10s %8.3f\n", fermenter->chiller_address, fermenter->chiller_state, fermenter->chiller_temperature);
	if (fermenter->heater_address) 
	    printf("Heater    %-36s %9d%% %8lu\n", fermenter->heater_address, fermenter->heater_state, fermenter->heater_usage);
	if (fermenter->cooler_address) 
	    printf("Cooler    %-36s %9d%% %8lu\n", fermenter->cooler_address, fermenter->cooler_state, fermenter->cooler_usage);
	if (fermenter->fan_address) 
	    printf("Fan       %-36s %9d%% %8lu\n", fermenter->fan_address, fermenter->fan_state, fermenter->fan_usage);
	if (fermenter->light_address) 
	    printf("Light     %-36s %9d%% %8lu\n", fermenter->light_address, fermenter->light_state, fermenter->light_usage);
	if (fermenter->door_address) 
	    printf("Door      %-36s %10d\n", fermenter->door_address, fermenter->door_state);
	if (fermenter->psu_address)
	    printf("PSU       %-36s %10d\n", fermenter->psu_address, fermenter->psu_state);
	printf("mode      %s\n", fermenter->mode);
	printf("alarm     %04x\n", fermenter->alarm);
	printf("sp high   %8.3f\n", fermenter->setpoint_high);
	printf("sp low    %8.3f\n", fermenter->setpoint_low);
	if (fermenter->profile_uuid) {
	    printf("profile uuid     %s\n", fermenter->profile_uuid);
	    printf("profile name     %s\n", fermenter->profile_name);
	    printf("profile state    %s\n", fermenter->profile_state);
	    printf("profile percent  %d\n", fermenter->profile_percent);
	    printf("profile sp high  %.3f\n", fermenter->profile_inittemp_high);
	    printf("profile sp low   %.3f\n", fermenter->profile_inittemp_low);
	    printf("profile steps    %s\n", fermenter->profile_steps);
	}
    }
}



void fermenter_death(char *topic)
{
    char		*edge_node, *alias, *msg = NULL;
    sys_fermenter_list	*tmpp;

    printf("fermenter_death: %s\n", topic);
    strtok(topic, "/"); // ignore namespace
    strtok(NULL, "/");	// ignore group_id
    strtok(NULL, "/");	// ignore message_type
    edge_node = strtok(NULL, "/\0");
    alias = strtok(NULL, "/\0");

    fermenter_mysql_death(edge_node, alias);

    if (alias) {
    	for (tmpp = fermenters; tmpp; tmpp = tmpp->next) {
	    if ((strcmp(tmpp->node, edge_node) == 0) && (strcmp(tmpp->alias, alias) == 0)) {
		if (tmpp->online) {
		    syslog(LOG_NOTICE, "Offline fermenter %s/%s", tmpp->node, tmpp->alias);
	    	    tmpp->online = false;
		    msg = xstrcpy((char *)"{\"device\":\"fermenters\",\"node\":\"");
    		    msg = xstrcat(msg, edge_node);
    		    msg = xstrcat(msg, (char *)"\",\"unit\":\"");
    		    msg = xstrcat(msg, alias);
    		    msg = xstrcat(msg, (char *)"\",\"online\":0}");
		    ws_broadcast(msg);
    		    free(msg);
    		    msg = NULL;
		}
	    	break;
	    }
    	}
    } else {
	for (tmpp = fermenters; tmpp; tmpp = tmpp->next) {
	    if (strcmp(tmpp->node, edge_node) == 0) {
		if (tmpp->online) {
		    syslog(LOG_NOTICE, "Offline fermenter %s/%s", tmpp->node, tmpp->alias);
		    tmpp->online = false;
		    msg = xstrcpy((char *)"{\"device\":\"fermenters\",\"node\":\"");
                    msg = xstrcat(msg, edge_node);
                    msg = xstrcat(msg, (char *)"\",\"unit\":\"");
                    msg = xstrcat(msg, tmpp->alias);
                    msg = xstrcat(msg, (char *)"\",\"online\":0}");
                    ws_broadcast(msg);
                    free(msg);
                    msg = NULL;
		}
	    }
	}
    }
}


